/* factorize.c - sieve of erathostenes */

#include <stdio.h>
#include <malloc.h>
#include <string.h>

static unsigned long *sieve = 0;
static int sieve_max = 0;
static int sieve_top = 0;

/*

divisible by 2 not represented
a='divisible by 3'
b='divisible by 5'
X='divisible by both 3 and 5'
(ab.a..X..a.ba..)*
(.ab.a..X..a.ba.)*

*/

/*static int sieve_map[15] = { -1,-1,0,-1,1,2,-1,3,4,-1,5,-1,-1,6,7 };*/
static int sieve_map[15] = { 0,-1,-1,1,-1,2,3,-1,4,5,-1,6,-1,-1,7 };
static int sieve_rmap[8] = { 0,3,5,6,8,9,11,14 };

static inline int sieve_index(int here)
{
	int map;
	if (here <= 1 || (here & 1) == 0) return -1;
	here >>= 1;
	map = sieve_map[here % 15];
	if (map < 0) return map;
	map += (here/15) << 3;
	if (map >= sieve_max) return -2;
	return map;
}

static inline int sieve_here(int idx)
{
	return ((((idx>>3)*15 + sieve_rmap[idx&7]) << 1) | 1);
}

static inline void sieve_mark(int here)
{
#if 1
	int there = here * here;
#else
	int there = here << 1;
#endif
	int idx;
	while (-2 != (idx = sieve_index(there))) {
		there += here;
		if (idx < 0) continue;
		sieve[idx>>5] |= (1<<(idx&31));
	}
}

static inline int sieve_next_idx(int idx)
{
	/* assert(idx >= 0); */
	while (idx < sieve_max && 0 != (sieve[idx>>5] & (1<<(idx&31)))) {
		idx++;
	}
	return idx;
}

static inline int sieve_next(int here)
{
	int idx;
	do {
		if (-2 == (idx = sieve_index(++here))) return -1;
	} while (idx < 0);
	return sieve_here(sieve_next_idx(idx));
}

static inline int sieve_get(int here)
{
	int idx = sieve_index(here);
	return (idx < 0) ? 1 : (0 != (sieve[idx>>5] & (1<<(idx&31))));
}

static inline int sieve_set(int here)
{
	int idx = sieve_index(here);
	if (idx >= 0) { sieve[idx>>5] |= (1<<(idx&31)); }
	return 1;
}

void alloc_sieve(int max)
{
	int size, bytes;

	sieve_max = max;
	while (-1 == (size = sieve_index(max))) max++;
	bytes = 4 + (size >> 5) * 4;
	sieve = malloc(bytes);
	if (!sieve) exit(1);
	sieve_max = size;
	sieve_top = max;
	memset(sieve, 0, bytes);
	printf("primes up to %d, sieve size %d\n", max, bytes);
}

void genprimes(char *max)
{
	int size;
	int here;
	int idx;
	char *p = max + strspn(max, "0123456789");
	size = atoi(max);
	switch(*p) {
		case 'm': case 'M': size *= 1024;
		case 'k': case 'K': size *= 1024; break;
	}
	alloc_sieve(size);
#if 1
	idx = sieve_index(7);
	do {
		sieve_mark(sieve_here(idx));
		idx = sieve_next_idx(++idx);
	} while (idx < sieve_max);
#else
	for (here=7; here!=-1; here=sieve_next(here)) {
		sieve_mark(here);
	}
#endif
}

void showprimes(void)
{
	int here;

	printf("2\n3\n5\n");
	for (here=7; here!=-1; here=sieve_next(here))
		printf("%d\n", here);
}

void factorize(int n)
{
	int factor;
	int cnt=0;

	printf("%d = ", n);
	for (factor=2; n>1 && factor!=-1; factor=sieve_next(factor))
		while (n == (n/factor)*factor) {
			printf(cnt++?"*%d":"%d", factor);
			n/=factor;
		}
	if (n>1)
		printf(cnt?"*(%d)":"(%d)", n);
	printf("\n");
}

#include <getopt.h>

int main(int argc, char **argv)
{
	char *max = "1000";
	int c;
	int quiet = 0;

	while (EOF != (c = getopt(argc, argv, "m:q"))) switch(c) {
		case 'm': max = optarg; break;
		case 'q': quiet = 1; break;
	}
	genprimes(max);
	if (optind >= argc) {
		if (!quiet) showprimes();
		return 0;
	}
	while (optind < argc) {
		c = atoi(argv[optind++]);
		factorize(c);
	}
	return 0;
}
